# Copyright (c) 2013, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <http://www.gnu.org/licenses/>.


import numpy as np
from scipy import signal

__all__ = ['find_tone']


def find_tone(x_in, f, fs=1, start=0, N=None, adjust_N=False, window='hamming',
              output='complex'):
    """Extracts a tone out of a signal where it might be embedded in noise.

    This is a complicate function to extract information (amplitude and
    phase) about a tone embedded in noise. In spite of its sophistication
    this function may not always work correctly.

    Parameters
    ----------
    x_in : array of floats
        input signal
    f : float
        frequency of tone being looked for
    fs : float
        sample frequency in the input vector x_in.
        Defaults to 1.
    start : integer
        first input sample to consider.
        Defaults to 0.
    N : integer or None
        number of input samples to consider. If it is None, then the number
        of samples taken in consideration is autoadjusted, using all the
        available samples, starting from start.
        Defaults to None.
    adjust_N : bool
        whether to adjust the number of samples to consider, in order to
        get an integer number of periodos of the tone being searched.
        Defaults to False.
    window : string, float or tuple
        type of window function used before practicing a Fourier Transform
        to look for the tone. Can be any window supported by the function
        get_window in the scipy signal module.
        Defaults to 'hanning'.
    output : string
        type of output to provide. Can be 'complex' or 'xy'.
        Defaults to 'complex'.

    Returns
    -------
    tone : complex or couple of floats.
        output format is controlled by input parameter `output`.
        If format is 'complex', then ouput encodes amplitude and phase of the
        input tone as a complex exponential. If format is 'xy' then real and
        'imaginary part' are provided separately.
    """
    M = len(x_in)
    fn = float(f)/fs
    samples_per_period = 1/fn
    Nmax = M-start
    if N is None:
        N = Nmax
    Pmax = int(np.floor(Nmax/samples_per_period))
    P = np.min((Pmax, int(np.round(N/samples_per_period))))
    if adjust_N:
        N2 = int(np.round(P*samples_per_period))
    else:
        N2 = int(np.min((Nmax, N)))
    if window == 'boxcar' or window is None:
        data = x_in[start:N2+start]
    else:
        win = signal.get_window(window, N2)
        adj = np.sum(win)/N2
        win = win/adj
        data = x_in[start:N2+start]*win
    x_cos = np.cos(2*np.pi*fn*np.arange(start, N2+start))
    x_sin = np.sin(2*np.pi*fn*np.arange(start, N2+start))
    b = 2*np.dot(x_cos, data)/N2
    a = 2*np.dot(x_sin, data)/N2
    if output == 'complex':
        return b+1j*a
    elif output == 'xy':
        return (b, a)
    else:
        return (np.sqrt(a**2+b**2), np.arctan2(a, b))



# -*- coding: utf-8 -*-

# Copyright (c) 2016, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <http://www.gnu.org/licenses/>.

"""
Monkey patch Sphinx, to use textgreek for Greek letters

* Restrict the tex_replacement table, to stop escaping Greek letters.
  * These are much better supported with the textgreek package.
  * Converting Greek text to math causes issues with hyperref when Greek
    letters occur in section headings.
  * Converting Greek text to math makes it impossible to make it follow
    the formatting of the surrounding text *eg. bold, italics, etc).
  * Note that this change also requires the unicode option to be passed to
    hyperref
"""

from __future__ import unicode_literals
from sphinx.util import texescape
from sphinx.writers import latex

greek_letters = 'αβγδεζηθικλμνξοπρςστυφχψωΑΒΓΔΕΖΗΘΙΚΛΜΝΞΟΠΡΣΤΥΦΧΨΩ'


def setup(app):
    tex_replacements_dict = dict(texescape.tex_replacements)
    for letter in greek_letters:
        if letter in tex_replacements_dict:
            del tex_replacements_dict[letter]
    texescape.tex_replacements = tex_replacements_dict.items()
#    latex.DEFAULT_SETTINGS['textgreek'] = (
#        '\\usepackage[artemisia]{textgreek}')
#    latex.HEADER = (latex.HEADER
#                    .replace("%(usepackages)s",
#                             "%(textgreek)s\n" +
#                             "%(usepackages)s")
#                    .replace("\\usepackage{sphinx}",
#                             "\\usepackage{sphinx}\n" +
#                             "\\hypersetup{unicode}"))
    return {'version': '0.1'}



# Copyright © 2013–2025 Sergio Callegari, Federico Bizzarri
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <http://www.gnu.org/licenses/>.

"""Demo for LP ΔΣ modulator with psychoacoustic noise weighting.

© 2013–2025 Sergio Callegari, Federico Bizzarri
All rights reserved.
"""


import numpy as np
import matplotlib.pyplot as plt
from pydsm.delsig import dbv, dbp
from pydsm.NTFdesign import ntf_dunn, ntf_fir_audio_weighting
from pydsm.delsig import simulateDSM, evalTF
import matplotlib.mlab as mlab
from pydsm.audio_weightings import f_weighting

# Signal specification
fsig = 1000.
B = 20500.
osr = 64
fphi = B*osr*2
# Lee constraint
H_inf = 1.5
# FIR Order
order = 30

# Computing the optimal NTF and the reference NTF
print("... computing optimal NTF")
# Reducing the feastol may help the optimization converge.
# Until reltol is very strict, subobtima are often returned
# Using a large normalization constant may help getting closer to the real
# optimum but may also hinder convergence
opti_ntf = ntf_fir_audio_weighting(order, osr, f_weighting,
                                   H_inf=H_inf,
                                   normalize=1E3,
                                   cvxopt_opts={'reltol': 1E-12,
                                                'abstol': 1E-10,
                                                'feastol': 1E-2})

dunn_ntf = ntf_dunn(3, osr, H_inf)

fmin = 10
fmax = fphi/2
ff = np.logspace(np.log10(fmin), np.log10(fmax), 1000)

resp_w = f_weighting(ff)
resp_opti = np.abs(evalTF(opti_ntf, np.exp(1j*2*np.pi*ff/fphi)))
resp_ref = np.abs(evalTF(dunn_ntf, np.exp(1j*2*np.pi*ff/fphi)))

# First figure. This provides the weighting function and the NTFs
fig0 = plt.figure()
plt.plot(ff/fphi, dbp(resp_w), 'b', label='audio weighting')
plt.plot(ff/fphi, dbv(resp_opti), 'r', label='proposed NTF')
l = plt.plot(ff/fphi, dbv(resp_ref), 'g', label="Dunn's NTF")
plt.suptitle('Audio weighting and NTF magnitude response')
plt.xlim(10E-5, 1./2)
plt.ylim(-140, 20)
plt.xscale('log', base=10)
plt.gca().set_xticks([1E-5, 1./2])
plt.gca().set_xticklabels(['$10^{-5}$', r'$\frac{1}{2}$'])
plt.xlabel('$f$', x=1.)
plt.ylabel((r'$w(f)$, ' +
            r'$\left|\mathit{NTF}\,\left(\mathrm{e}^{\mathrm{i} 2\pi f}' +
            r'\right)\right|$ [dB]'), y=0.5)
plt.grid(True, 'both')
plt.legend(loc='lower right')
plt.tight_layout(rect=[0, 0, 1, 0.98])

# Start and stop time for DS simulation
Tstop = 400000
Tstart = 40000
dither_sigma = 0
# Use very little signals, otherwise it is troublesome to
# remove them completely to evaluate the quantization noise
A = 0.01
DT = 4*fphi/fsig

# Setting up DS simulation
tt = np.asarray(range(int(Tstop)))
uu = A*np.sin(2*np.pi*fsig/fphi*tt)
dither = np.random.randn(len(uu))*dither_sigma
uud = uu+dither

# Simulating the DS
# Some of the commented bits are used to obtain the quantizazion noise pdf
print("Simulating optimal NTF")
xx_opti = simulateDSM(uud, opti_ntf)[0]
xx_ref = simulateDSM(uud, dunn_ntf)[0]

NFFT = 4096*32
# Possibly this can be improved by extracting the tone from the
# output to do the subtraction
(psd1, freqs) = mlab.psd((xx_opti-uu)[Tstart:Tstop], Fs=fphi, NFFT=NFFT,
                         noverlap=NFFT//2, scale_by_freq=True)
(psd2, reqs) = mlab.psd((xx_ref-uu)[Tstart:Tstop], Fs=fphi, NFFT=NFFT,
                        noverlap=NFFT//2, scale_by_freq=True)

minfreq = 1
df = freqs[1]-freqs[0]
print('frequency step in psd data=', df)

psd1 *= df
psd2 *= df

minidx = int(np.floor(minfreq/df))
print('minimum frequency index for plots', minidx)
maxidx = int(np.ceil(22E3/df))

# Here we plot the spectra obtained by time domain data
plt.figure()
plt.semilogx(freqs[minidx:], dbp(psd1[minidx:]), 'r', label='proposed')
plt.semilogx(freqs[minidx:], dbp(psd2[minidx:]), 'g', label="Dunn's",
             alpha=0.4)
plt.semilogx(ff, dbp(1.3*resp_opti**2/(fphi)), 'k', label='proposed NTF')
plt.legend(loc='lower right')
plt.title('Spectra from time domain data')
plt.tight_layout(rect=[0, 0, 1, 0.98])

ww = np.zeros_like(psd1)
ww = np.asarray([max(f_weighting(freq), 1E-200) for freq in freqs])

plt.figure()
plt.plot(freqs[minidx:maxidx], dbp((psd1*ww)[minidx:maxidx]), 'r',
         label='proposed')
l = plt.plot(freqs[minidx:maxidx], dbp((psd2*ww)[minidx:maxidx]), 'g',
             label="Dunn's")
l[0].set_dashes([2, 2])
plt.xscale('log', base=10)
plt.xlim(freqs[minidx], 22E3)
plt.ylim(-240, -120)
plt.xlabel('$f$', x=1.)
plt.ylabel('Loudness dB', y=0.5)
plt.grid(True, 'both')
plt.title("Perceived noise level")
plt.legend(loc='lower center')
plt.tight_layout(rect=[0, 0, 1, 0.98])

plt.show()



# Copyright © 2012–2025 Sergio Callegari, Federico Bizzarri
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <http://www.gnu.org/licenses/>.

"""Demo code to replicate the result shown in Figure 4 of the paper:

S. Callegari 'Coding of Stereo Signals by a Single Digital ΔΣ Modulator'
IEEE 20th International Conference on Electronics, Circuits, and Systems
(ICECS 2013), pp. 589–592, Dec. 2013.

© 2012–2025 Sergio Callegari, Federico Bizzarri
All rights reserved.
"""

# Note: code rises a warning since in one of the cases being tested
# the NTF synthesis function is invoked with a too large maximum peak
# gain (Lee constraint) together with a very low modulator order.
# In this conditions, the synthesised NTF fails to reach the max gain.
# This does not affect the visualized trends, though.


import numpy as np
from pydsm.delsig import (synthesizeNTF, ds_f1f2, dbp)
from pydsm.NTFdesign.weighting import quantization_weighted_noise_gain
import matplotlib.pyplot as plt

# Set the OSRs and the h_inf (gamma) values used for the plots
# OSR values are arranged so that each one is twice the previous one
# hinf values are arranged so that each value is the square root of the
# previous one. Also set the orders used for testing.
osrs = np.asarray([16, 32, 64, 128, 256])
hinfs = np.asarray([2.25**(1/(2.**i)) for i in range(5)])
orders = np.asarray([1, 2, 3, 4, 5])

# Prepare a vector to store the noise gain for obtained at the many
# test conditions
g0s = np.zeros((orders.size, osrs.size))

# Synthesize the NTF for all the orders and OSR values at h_inf=1.5
for i, osr in enumerate(osrs):
    for j, order in enumerate(orders):
        ntf1 = synthesizeNTF(order, osr=osr, opt=3, H_inf=1.5)
        f1, f2 = ds_f1f2(osr)
        g0 = quantization_weighted_noise_gain(ntf1, None, (f1, f2))
        # print order, osr, f1, f2, g0, dbp(g0)
        g0s[j, i] = g0
# Make the plot: quantization noise gain versus OSR for the different
# orders
plt.figure()
plt.xscale('log')
markers = 'ov^<>'
for i, order in enumerate(orders):
    plt.plot(osrs, dbp(g0s[i]), "-"+markers[i], label='%d' % order)
plt.legend(loc='upper right', fontsize=9)
plt.xlabel("OSR")
plt.ylabel("$P_N$ [dB]")
# Print some interesting data from the designs
print("Analysis of data for varying OSR")
for i, order in enumerate(orders):
    print("Order: %d. Average slope: %f dB per OSR doubling." %
          (order, (dbp(g0s[i, -1])-dbp(g0s[i, 0]))/(orders.size-1)))

# Prepare a vector to store the noise gain for obtained at the many
# test conditions
g0s = np.zeros((orders.size, hinfs.size))

# Synthesize the NTF for all the orders and h_inf values at OSR=64
for i, hinf in enumerate(hinfs):
    for j, order in enumerate(orders):
        ntf1 = synthesizeNTF(order, osr=64, opt=3, H_inf=hinf)
        f1, f2 = ds_f1f2(osr)
        g0 = quantization_weighted_noise_gain(ntf1, None, (f1, f2))
        # print order, hinf, f1, f2, g0, dbp(g0)
        g0s[j, i] = g0
# Make the plot: quantization noise gain versus h_inf for the different
# orders
plt.figure()
markers = 'ov^<>'
for i, order in enumerate(orders):
    plt.plot(np.arange(5)+1, dbp(g0s[i]), "-"+markers[i], label='%d' % order)
plt.legend(loc='lower right', fontsize=9)
plt.xlabel(r"$n$ so that $\gamma = \gamma_0^{\frac{1}{2^n}}$")
plt.ylabel("$P_N$ [dB]")
# Print some interesting data from the designs
print("\nAnalysis of data for varying h_inf")
for i, order in enumerate(orders):
    print("Order: %d. Average slope: %f dB per square root of h_inf." %
          (order, (dbp(g0s[i, 2])-dbp(g0s[i, 1]))))

plt.show()



# -*- coding: utf-8 -*-

# Copyright (c) 2016, Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <http://www.gnu.org/licenses/>.

# This file includes code from Sphinx 1.3.5
# covered by the following copyright and permission notice

# Copyright (c) 2007-2016 by the Sphinx team
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Monkey patch Sphinx, in order to use the ltytable package

* This is to provide better support support for multi page tables.
  * The ltytable package is not an official ctan package, rather an attempt
    at mixing tabulary and longtable, based on a proposal by David Carlise.
    see http://tex.stackexchange.com/questions/78075/multi-page-with-tabulary
  * Namely use the longtabu environment in place of longable.
  * The new ltabulary environment supports tabulary features such as the
    possibility of having automatically sized paragraph type columns.  This
    is needed by autosummary tables to avoid overfull boxes in the
    function descriptions.

* Also have autosummary use the new column types
"""

from __future__ import unicode_literals
import os
from os import path
from sphinx.writers import latex
from sphinx.builders.latex import LaTeXBuilder
from sphinx.locale import _
from sphinx.ext.autosummary import autosummary_table, Autosummary
from sphinx import addnodes
from docutils import nodes
from docutils.statemachine import ViewList
from sphinx.util.console import bold
from sphinx.util.osutil import copyfile

latex_package_path = (path.dirname(path.realpath(__file__)) +
                      '/../texinputs/')

old_finish = LaTeXBuilder.finish


def new_finish(self):
    old_finish(self)
    self.info(bold('copying more TeX support files...'))
    for filename in os.listdir(latex_package_path):
        if not filename.startswith('.'):
            copyfile(path.join(latex_package_path, filename),
                     path.join(self.outdir, filename))


def depart_table(self, node):
    if self.table.rowcount > 30:
        self.table.longtable = True
    self.popbody()
    if not self.table.longtable and self.table.caption is not None:
        self.body.append('\n\n\\begin{threeparttable}\n'
                         '\\capstart\\caption{')
        for caption in self.table.caption:
            self.body.append(caption)
        self.body.append('}')
        for id in self.next_table_ids:
            self.body.append(self.hypertarget(id, anchor=False))
        if node['ids']:
            self.body.append(self.hypertarget(node['ids'][0],
                                              anchor=False))
        self.next_table_ids.clear()
    if self.table.longtable:
        self.body.append('\n\\begin{ltabulary}')
        endmacro = '\\end{ltabulary}\n\n'
    elif self.table.has_verbatim:
        self.body.append('\n\\begin{tabular}')
        endmacro = '\\end{tabular}\n\n'
    elif self.table.has_problematic and not self.table.colspec:
        # if the user has given us tabularcolumns, accept them and use
        # tabulary nevertheless
        self.body.append('\n\\begin{tabular}')
        endmacro = '\\end{tabular}\n\n'
    else:
        self.body.append('\n\\begin{tabulary}{\\linewidth}')
        endmacro = '\\end{tabulary}\n\n'
    if self.table.colspec:
        self.body.append(self.table.colspec)
    else:
        if self.table.has_problematic:
            colwidth = 0.95 / self.table.colcount
            colspec = ('p{%.3f\\linewidth}|' % colwidth) * \
                self.table.colcount
            self.body.append('{|' + colspec + '}\n')
        elif self.table.longtable:
            self.body.append('{|' + ('l|' * self.table.colcount) + '}\n')
        else:
            self.body.append('{|' + ('L|' * self.table.colcount) + '}\n')
    if self.table.longtable and self.table.caption is not None:
        self.body.append(u'\\caption{')
        for caption in self.table.caption:
            self.body.append(caption)
        self.body.append('}')
        for id in self.next_table_ids:
            self.body.append(self.hypertarget(id, anchor=False))
        self.next_table_ids.clear()
        self.body.append(u'\\\\\n')
    if self.table.longtable:
        self.body.append('\\hline\n')
        self.body.extend(self.tableheaders)
        self.body.append('\\endfirsthead\n\n')
        self.body.append('\\multicolumn{%s}{c}%%\n' % self.table.colcount)
        self.body.append(r'{{\textsf{\tablename\ \thetable{} -- %s}}} \\'
                         % _('continued from previous page'))
        self.body.append('\n\\hline\n')
        self.body.extend(self.tableheaders)
        self.body.append('\\endhead\n\n')
        self.body.append((r'\hline \multicolumn{%s}{|r|}{{\textsf{%s}}}' +
                          r'\\ \hline')
                         % (self.table.colcount,
                            _('Continued on next page')))
        self.body.append('\n\\endfoot\n\n')
        self.body.append('\\endlastfoot\n\n')
    else:
        self.body.append('\\hline\n')
        self.body.extend(self.tableheaders)
    self.body.extend(self.tablebody)
    self.body.append(endmacro)
    if not self.table.longtable and self.table.caption is not None:
        self.body.append('\\end{threeparttable}\n\n')
    self.unrestrict_footnote(node)
    self.table = None
    self.tablebody = None


def get_table(self, items):
    """Generate a proper list of table nodes for autosummary:: directive.

    *items* is a list produced by :meth:`get_items`.
    """
    table_spec = addnodes.tabular_col_spec()
    table_spec['spec'] = 'lL'

    table = autosummary_table('')
    real_table = nodes.table('', classes=['longtable'])
    table.append(real_table)
    group = nodes.tgroup('', cols=2)
    real_table.append(group)
    group.append(nodes.colspec('', colwidth=10))
    group.append(nodes.colspec('', colwidth=90))
    body = nodes.tbody('')
    group.append(body)

    def append_row(*column_texts):
        row = nodes.row('')
        for text in column_texts:
            node = nodes.paragraph('')
            vl = ViewList()
            vl.append(text, '<autosummary>')
            self.state.nested_parse(vl, 0, node)
            try:
                if isinstance(node[0], nodes.paragraph):
                    node = node[0]
            except IndexError:
                pass
            row.append(nodes.entry('', node))
        body.append(row)

    for name, sig, summary, real_name in items:
        qualifier = 'obj'
        if 'nosignatures' not in self.options:
            col1 = ':%s:`%s <%s>`\ %s' % (qualifier, name, real_name, sig)
        else:
            col1 = ':%s:`%s <%s>`' % (qualifier, name, real_name)
        col2 = summary
        append_row(col1, col2)

    return [table_spec, table]


def setup(app):
    # First of all, assure that the latex package files are copyed to the
    # latex build directory
#    LaTeXBuilder.finish = new_finish
#    # Then, assure that the ltytable package is loaded in place of the
#    # longtable package
#    latex.LaTeXTranslator.default_elements['ltytable'] = (
#        '\\usepackage{ltytable}')
#    latex.HEADER = (latex.HEADER
#                    .replace("%(longtable)s",
#                             "%(ltytable)s"))
#    # And use ltabulary in place of longtable
#    latex.LaTeXTranslator.depart_table = depart_table
#    # also redefine the get_table func in autosummary
#    Autosummary.get_table = get_table
    return {'version': '0.1'}



#***********************************************************************#
# Copyright (C) 2010-2012 Tomas Tinoco De Rubira                        #
#                                                                       #
# This file is part of CVXPY                                            #
#                                                                       #
# CVXPY is free software: you can redistribute it and/or modify         #
# it under the terms of the GNU General Public License as published by  #
# the Free Software Foundation, either version 3 of the License, or     #
# (at your option) any later version.                                   #
#                                                                       #
# CVXPY is distributed in the hope that it will be useful,              #
# but WITHOUT ANY WARRANTY; without even the implied warranty of        #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
# GNU General Public License for more details.                          #
#                                                                       #
# You should have received a copy of the GNU General Public License     #
# along with this program.  If not, see <http://www.gnu.org/licenses/>. #
#***********************************************************************#

import numpy as np
from ..defs import *
from ..utils import *
from ..interface import *
from ..functions.abs import abs
from ..arrays import cvxpy_matrix

# Norm infinity
def norm_inf(X):
    r"""
    | :math:`\mbox{norm\_inf} :
      \mathbb{R}^{m \\times n} \\to \mathbb{R},
      \ \mbox{norm\_inf}(X) = \mbox{sup} \{ \|Xu\|_{\infty} \ |
      \ \|u\|_{\infty} \leq 1 \}`.
    | Convex.

    :param X: number,
              :ref:`scalar object<scalar_ref>` or
              :ref:`multidimensional object<multi_ref>`.
    :return: number
             or :ref:`tree<tree_obj>`.
    """

    # Check input
    if (np.isscalar(X) or
        type(X).__name__ in SCALAR_OBJS):
        X = vstack([X])
    elif (type(X) is not cvxpy_matrix and
          type(X).__name__ not in ARRAY_OBJS):
        raise TypeError('Invalid argument type')

    # Construct program
    (m,n) = X.shape
    t = variable()
    z = variable(m,n)
    w = vstack([z])
    p = program(minimize(t),
                list(map(lambda y: leq(sum(abs(y)),t),
                    [w[i,:] for i in range(0,m)])),
                [z],
                name='norm_inf')
    return p(X)



# Copyright © 2013–2025 Sergio Callegari, Federico Bizzarri
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <http://www.gnu.org/licenses/>.

"""Demo for ΔΣ modulator with first order LP filter.

© 2013–2025 Sergio Callegari, Federico Bizzarri
All rights reserved.
"""

# Note. This code returns numeric results that are not fully consistent with
# those in the TCAS1 2013 paper. Specifically, the SNR values
# evaluated via the NTF on the paper are better than those obtained by
# this code by a factor 2. In previous versions of the code, the results
# where identical due to a missing factor 2 in the quantization noise gain
# evaluation. Probably, this error slipped in because it makes the NTF-based
# results and the 'real' time-domain-simulation results more similar. In fact,
# the 'real' modulator behavior is better than predicted by the NTF model
# because in the specific example the quantization noise is not exactly white,
# but slightly blue (for some reason it gets more power at higher frequencies).


import numpy as np
import scipy as sp
import matplotlib.pyplot as plt
from pydsm.ir import impulse_response
from pydsm.delsig import synthesizeNTF, simulateDSM, evalTF, dbv, dbp
from pydsm.NTFdesign import quantization_noise_gain
from pydsm.NTFdesign.legacy import q0_from_filter_ir
from pydsm.NTFdesign.weighting import ntf_fir_from_q0

# Signal specification
fsig = 1000.
B = 1000.
OSR = 1024
fphi = B*OSR*2
# Lee constraint
H_inf = 1.5
# FIR Order for optimal NTF
order = 12
# Signal amplitude
A = 0.4

# Generate filter. Transfer function is normalized to be 0dB in pass band
# As an example, take cutoff freq at twice the top of the signal band to avoid
# attenuation when the input signal is close to it.
print("...generating filter")
# Care: in butter the cut of frequency is specified as a number from 0 to 1
# where 1 is fphi/2, not fphi
hz = sp.signal.butter(1, 2*(2*B)/fphi, btype='low')

# Compute impulse response
print("...computing impulse response of filter")
hz_ir = impulse_response(hz, db=60)

# Compute the optimal NTF
print("... computing optimal NTF")
q0 = q0_from_filter_ir(order, hz_ir)
ntf_opti = ntf_fir_from_q0(q0, H_inf=H_inf)

# Compute an NTF with DELSIG, for comparison
print("... computing delsig NTF")
ntf_delsig = synthesizeNTF(4, OSR, 3, H_inf, 0)

# Determine freq values for which plots are created
fmin = 10**np.ceil(np.log10(2*B/OSR))
fmax = 10**np.floor(np.log10(fphi/2))
ff = np.logspace(np.log10(fmin), np.log10(fmax), 1000)

# Compute frequency response data
resp_filt = np.abs(evalTF(hz, np.exp(1j*2*np.pi*ff/fphi)))
resp_opti = np.abs(evalTF(ntf_opti, np.exp(1j*2*np.pi*ff/fphi)))
resp_delsig = np.abs(evalTF(ntf_delsig, np.exp(1j*2*np.pi*ff/fphi)))

# Plot frequency response
plt.figure()
plt.semilogx(ff, dbv(resp_filt), 'b', label="Output filter")
plt.semilogx(ff, dbv(resp_opti), 'r', label="Optimal NTF")
plt.semilogx(ff, dbv(resp_delsig), 'g', label="Delsig NTF")
plt.legend(loc="lower right")
plt.suptitle("Output filter and NTFs")

# Check merit factors
ffl = np.linspace(fmin, fmax, 1000)
pg_opti = (np.abs(evalTF(ntf_opti, np.exp(1j*2*np.pi*ffl/fphi))) *
           np.abs(evalTF(hz, np.exp(1j*2*np.pi*ffl/fphi))))
pg_delsig = (np.abs(evalTF(ntf_delsig, np.exp(1j*2*np.pi*ffl/fphi))) *
             np.abs(evalTF(hz, np.exp(1j*2*np.pi*ffl/fphi))))
plt.figure()
plt.plot(ffl, pg_opti**2, 'r', label="Optimal NTF")
plt.plot(ffl, pg_delsig**2, 'g', label="Delsig NTF")
plt.legend(loc="upper right")
plt.suptitle("Merit factor integrand")

# Compute expected behavior
sigma2_e = 1./3
noise_power_opti_1 = quantization_noise_gain(hz, ntf_opti)*sigma2_e
noise_power_delsig_1 = quantization_noise_gain(hz, ntf_delsig)*sigma2_e
print("Expected optimal noise level {} ({} dB).\nExpected SNR {} dB".format(
    noise_power_opti_1, dbp(noise_power_opti_1),
    dbp(0.5*A**2)-dbp(noise_power_opti_1)))
print("Expected delsig noise level {} ({} dB).\nExpected SNR {} dB".format(
    noise_power_delsig_1, dbp(noise_power_delsig_1),
    dbp(0.5*A**2)-dbp(noise_power_delsig_1)))

# Start and stop time for DS simulation
Tstop = 100000
Tstart = 40000
dither_sigma = 1e-6

# Set up DSM simulation
tt = np.asarray(range(int(Tstop)))
uu = A*np.sin(2*np.pi*fsig/fphi*tt)
dither = np.random.randn(len(uu))*dither_sigma
uud = uu+dither

# Simulate the DSM
print("Simulating optimal NTF")
xx_opti = simulateDSM(uud, ntf_opti)[0]
print("Simulating delsig NTF")
xx_delsig = simulateDSM(uud, ntf_delsig)[0]

print("Applying the output filter")
uu_filt = sp.signal.lfilter(hz[0], hz[1], uu)
xx_opti_filt = sp.signal.lfilter(hz[0], hz[1], xx_opti)
xx_delsig_filt = sp.signal.lfilter(hz[0], hz[1], xx_delsig)

plt.figure()
plt.plot(tt[Tstart:Tstart+4*OSR], uu_filt[Tstart:Tstart+4*OSR], 'b',
         label="Filtered input")
plt.plot(tt[Tstart:Tstart+4*OSR], xx_opti_filt[Tstart:Tstart+4*OSR], 'r',
         label="Filtered Optimal DSM output")
plt.plot(tt[Tstart:Tstart+4*OSR], xx_delsig_filt[Tstart:Tstart+4*OSR], 'g',
         label="Filtered Delsig DSM output")
plt.legend(loc="best")
plt.suptitle("Portion of time domain behavior")

print("Filtering input")
uud_filt = sp.signal.lfilter(hz[0], hz[1], uud)

noise_power_opti_2 = (np.sum((uud_filt[Tstart:]-xx_opti_filt[Tstart:])**2) /
                      (Tstop-Tstart))
noise_power_delsig_2 = (np.sum((uud_filt[Tstart:] -
                                xx_delsig_filt[Tstart:])**2) /
                        (Tstop-Tstart))

print("Observed optimal noise level {} ({} dB)\nObserved SNR {} dB".format(
    noise_power_opti_2, dbp(noise_power_opti_2),
    dbp(0.5*A**2)-dbp(noise_power_opti_2)))
print("Observed delsig noise level {} ({} dB)\nObserved SNR {} dB".format(
    noise_power_delsig_2, dbp(noise_power_delsig_2),
    dbp(0.5*A**2)-dbp(noise_power_delsig_2)))

plt.show()



# Copyright © 2013-2025 Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <http://www.gnu.org/licenses/>.


import numpy as np
from scipy.signal import tf2zpk, zpk2tf

__all__ = ['mirroredNTF']


def mirroredNTF(ntf):
    """Creates a symmetric noise transfer function from a prototype.

    This function takes a prototype single band NTF for a LP modulator
    and returns a symmetric NTF with signal bands at the bottom and at
    the top of the available frequency ranges.

    Note that the way in which the mirroring happens make the resulting
    NTF peak at about the squared value as the peak of the original
    prototype. Hence the prototype should be designed with an H_inf value
    that is the square root of the desired one.

    Parameters
    ----------
    ntf : tuple
        input noise tranfer function in zpk form

    Returns
    -------
    ntf2 : tuple
        output noise transfer function in zpk form
    """
    zpk_mode = True
    if len(ntf) != 3:
        zpk_mode = False
        ntf = tf2zpk(*ntf)
    # Take the opposite of poles and zeros to get an HP filter
    ntf_flipped = (-ntf[0], -ntf[1], ntf[2])
    # Take the product of the two
    ntf_mirrored = (np.hstack((ntf[0], ntf_flipped[0])),
                    np.hstack((ntf[1], ntf_flipped[1])), 1)
    if not zpk_mode:
        ntf_mirrored = zpk2tf(*ntf_mirrored)
    return ntf_mirrored



# Copyright © 2013–2025 Sergio Callegari, Federico Bizzarri
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <http://www.gnu.org/licenses/>.

"""Demo for LP ΔΣ modulator with low DC noise.

© 2013–2025 Sergio Callegari, Federico Bizzarri
All rights reserved.
"""


import numpy as np
from numpy.random import randn
import matplotlib.pyplot as plt
import matplotlib.mlab as mlab
from pydsm.NTFdesign import ntf_fir_weighting
from pydsm.delsig import evalTF, dbv, dbp, simulateDSM

# Signal specification
fsig = 500.
B = 1000.
OSR = 64
fphi = B*OSR*2
# Lee constraint
H_inf = 1.5
# FIR Order for optimal NTF
order = 32


# Define the weighting function
# Start with a brickwall Lp function cutting at B
def w1(f):
    return 1. if f < B/fphi else 1e-12


# Introduce a weighting component that magnifies weight for
# frequencies < B/10
def weight2(f):
    return 50.-49.*f/(B/fphi/10.) if f < B/fphi/10. else 1.


# Get the final weighting function as the product of the two
def w2(f):
    return w1(f)*weight2(f)


print("... computing optimal NTF...")
ntf1 = ntf_fir_weighting(order, w1, quad_opts={'points': [B/fphi]})
ntf2 = ntf_fir_weighting(order, w2, quad_opts={'points': [B/fphi, B/fphi/10]})

# Prepare frequency axis for plotting
fmin = 10**np.ceil(np.log10(B/OSR/100))
fmax = fphi/2
ff = np.logspace(np.log10(fmin), np.log10(fmax), 1000)

resp_w1 = np.fromiter(map(w1, ff/fphi), np.double)
resp_ntf1 = np.abs(evalTF(ntf1, np.exp(1j*2*np.pi*ff/fphi)))
resp_w2 = np.fromiter(map(w2, ff/fphi), np.double)
resp_ntf2 = np.abs(evalTF(ntf2, np.exp(1j*2*np.pi*ff/fphi)))

ffa = np.logspace(np.log10(0.5E-5), np.log10(0.5), 1024)
vv_a2 = dbv(np.abs(evalTF(ntf2, np.exp(1j*2*np.pi*ffa))))

fig0 = plt.figure()
l = plt.plot(ff/fphi, dbp(resp_w1), 'b', label='on-off weighting')
plt.plot(ff/fphi, dbp(resp_w2), 'r', label='low-dc-noise weighting')
l = plt.plot(ff/fphi, dbv(resp_ntf1), 'b--', label='NTF - on-off')
plt.plot(ff/fphi, dbv(resp_ntf2), 'r--', label='NTF - low-dc-noise')
plt.xlim(1e-5, 1./2)
plt.xscale('log', base=10)
plt.gca().set_xticks([1E-5, 1./2])
plt.gca().set_xticklabels(['$10^{-5}$', r'$\frac{1}{2}$'])
plt.xlabel('$f$', x=1.)
plt.ylabel((r'$w(f)$, ' +
            r'$\left|\mathit{NTF}\,\left(\mathrm{e}^{\mathrm{i} 2\pi f}' +
            r'\right)\right|$ [dB]'), y=0.5)
plt.grid(True, 'both')
plt.suptitle('Noise weighting and NTF magnitude response')
plt.legend(loc='lower right')
plt.tight_layout(rect=[0, 0, 1, 0.98])

# Create an input signal for simulation
# Two signal components
fa = 53.
fb = 497.
periods = 2000
nfft_periods = 40
skip_periods = 10
tmax = int(np.ceil(periods*1./fa*fphi))
tplot0 = int(np.round(skip_periods*1./fa*fphi))
tplot2 = int(np.round((periods-2)*1./fa*fphi))
nn = np.arange(tmax+1)

# Three signal components available. Play freely with DC, slow and fast
# amplitudes
# Plot in paper obtained with
# c1=0 c2=0 c3=0.11111*np.sin(2*np.pi*fb*nn/fphi)*1E-3
#
# Constant component
c1 = 0.11111*0
# Slow component
c2 = 0*np.sin(2*np.pi*fa*nn/fphi)*1E-3
# Fast component
c3 = 0.11111*np.sin(2*np.pi*fb*nn/fphi)*1E-3
# dither
dither = 2.5E-6*randn(tmax+1)
sig = c1+c2+c3+dither

print("Simulating DSM")
ds1 = simulateDSM(sig, ntf1)[0]
ds2 = simulateDSM(sig, ntf2)[0]

ndt = tplot2-tplot0+1
dcval1 = np.sum(ds1[tplot0:tplot2])/ndt
dcval2 = np.sum(ds2[tplot0:tplot2])/ndt
print('Accuracy in the recovery of DC components')
print('recovered DC on-off', dcval1)
print('recovered DC low-dc-noise', dcval2)
print('original DC', np.sum(sig[tplot0:tplot2])/ndt)

print("Computing spectra")
NFFT = int(2**(np.ceil(np.log(nfft_periods*1./fa*fphi)/np.log(2))))
# This is a bit rough... there may be some residual power at the
# signal frequencies. To obtain a good plot make c1, c2, c3 very small
(psd1, freqs) = mlab.psd((ds1-sig)[tplot0:tplot2], Fs=2*fphi, NFFT=NFFT,
                         noverlap=NFFT//2)
(psd2, freqs) = mlab.psd((ds2-sig)[tplot0:tplot2], Fs=2*fphi, NFFT=NFFT,
                         noverlap=NFFT//2)
minfreq = 1
df = freqs[1]-freqs[0]
minidx = int(np.ceil(minfreq/df))

plt.figure()
plt.plot(freqs[minidx:], dbp(psd1[minidx:]), 'b', label='on-off')
plt.plot(freqs[minidx:], dbp(psd2[minidx:]), 'r', label='low-dc-noise')
plt.xlim(2, 128E3)
plt.xscale('log', base=10)
plt.xlabel('$f$', x=1.)
plt.ylabel('Quant. noise PSD [dB]')
plt.grid(True, 'both')
plt.legend(loc="upper left")
plt.tight_layout(rect=[0, 0, 1, 0.98])

plt.show()



# Copyright © 2013–2025 Sergio Callegari, Federico Bizzarri
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <http://www.gnu.org/licenses/>.

"""Demo for multiband ΔΣ modulator with brickwall filter.

© 2013–2025 Sergio Callegari, Federico Bizzarri
All rights reserved.
"""


import numpy as np
from pydsm.delsig import evalTF
from pydsm.delsig import dbv
import matplotlib.pyplot as plt
from pydsm.NTFdesign import ntf_fir_weighting

import matplotlib
matplotlib.rc('font', size=16, family='serif')

# Signal specification
B = 1000.
OSR = 64
fphi = B*OSR*2
# Lee constraint
H_inf = 1.5
# FIR Order for optimal NTF
order = 32


# On-off weighting function
def w1(f):
    return 1. if f < B/fphi or 8*B/fphi < f < 15*B/fphi else 1e-12


print("... plotting filter magnitude response")
ff = np.logspace(np.log10(0.5E-5), np.log10(0.5), 1024).reshape((-1, 1))
vv = np.fromiter(map(w1, ff), np.double)
plt.figure()
plt.plot(ff, vv)
plt.xlim(1e-5, 1./2)
plt.ylim(0, 1.1)
plt.xscale('log', base=10)
plt.gca().set_xticks([1E-5, 1./2])
plt.gca().set_xticklabels(['$10^{-5}$', r'$\frac{1}{2}$'])
plt.xlabel('$f$', x=1.)
plt.ylabel('$w(f)$', y=0.9)
plt.grid(True, 'both')
plt.suptitle('Weighting function')
plt.tight_layout(rect=[0, 0, 1, 0.98])

print("... computing optimal NTF")
ntf_opti = ntf_fir_weighting(order, w1, H_inf=H_inf,
                             cvxopt_opts={'reltol': 1E-11,
                                          'abstol': 1E-9,
                                          'feastol': 1E-9})
ntf_opti_mag = lambda f: np.abs(evalTF(ntf_opti, np.exp(-2j*np.pi*f)))


print("... plotting optimal NTF amplitude response")
vv_mag = dbv(ntf_opti_mag(ff))
plt.figure()
plt.plot(ff, vv_mag)
plt.xlim(1e-5, 1./2)
# plt.ylim(0, 1.1)
plt.xscale('log', base=10)
plt.gca().set_xticks([1E-5, 1./2])
plt.gca().set_xticklabels(['$10^{-5}$', r'$\frac{1}{2}$'])
plt.xlabel('$f$', x=1.)
plt.ylabel((r'$\left|\mathit{NTF}\,\left(\mathrm{e}^{\mathrm{i} 2\pi f}' +
            r'\right)\right|$ [dB]'), y=0.59)
plt.grid(True, 'both')
plt.suptitle('Noise transfer function (magnitude)')
plt.tight_layout()


plt.show()



#***********************************************************************#
# Copyright (C) 2010-2012 Tomas Tinoco De Rubira                        #
#                                                                       #
# This file is part of CVXPY                                            #
#                                                                       #
# CVXPY is free software: you can redistribute it and/or modify         #
# it under the terms of the GNU General Public License as published by  #
# the Free Software Foundation, either version 3 of the License, or     #
# (at your option) any later version.                                   #
#                                                                       #
# CVXPY is distributed in the hope that it will be useful,              #
# but WITHOUT ANY WARRANTY; without even the implied warranty of        #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
# GNU General Public License for more details.                          #
#                                                                       #
# You should have received a copy of the GNU General Public License     #
# along with this program.  If not, see <http://www.gnu.org/licenses/>. #
#***********************************************************************#

import numpy as np
from ..defs import *
from ..utils import *
from ..sets import *
from ..interface import *
from ..arrays import cvxpy_matrix

# Kullback-Leibler divergence
def kl_div(x,y):
    r"""
    | :math:`\mbox{kl\_div} :
      \mathbb{R}_{++}^n \\times \mathbb{R}_{++}^n \\to \mathbb{R},
      \ \mbox{kl\_div}(x,y) = \sum_{i=1}^n
      ( x_i \mbox{log} (x_i/y_i) - x_i + y_i)`.
    | Convex.

    :param x: number,
              :ref:`scalar object<scalar_ref>` or
              :ref:`multidimensional object<multi_ref>`.
    :param y: number,
              :ref:`scalar object<scalar_ref>` or
              :ref:`multidimensional object<multi_ref>`.
    :return: number or
             :ref:`tree<tree_obj>`.
    """

    # Check x
    if (np.isscalar(x) or
        type(x).__name__ in SCALAR_OBJS):
        x = vstack([x])
    elif (type(x) is not cvxpy_matrix and
          type(x).__name__ not in ARRAY_OBJS):
        raise TypeError('Invalid first argument type')
    elif x.shape[1] != 1:
        raise ValueError('Invalid first argument dimension')

    # Check y
    if (np.isscalar(y) or
        type(y).__name__ in SCALAR_OBJS):
        y = vstack([y])
    elif (type(y) is not cvxpy_matrix and
          type(y).__name__ not in ARRAY_OBJS):
        raise TypeError('Invalid second argument type')
    elif y.shape[1] != 1:
        raise ValueError('Invalid second argument dimension')

    # Check dimension match
    if x.shape[0] != y.shape[0]:
        raise ValueError('Dimension mismatch')

    # Check positive if numeric
    n = x.shape[0]
    for i in range(0,n,1):
        if np.isscalar(x[i,0]) and x[i,0] <= 0:
            return np.inf
        if np.isscalar(y[i,0]) and y[i,0] <= 0:
            return np.inf

    # Construct objective and constraints
    t = variable()
    u = variable(n,1)
    v = variable(n,1)
    constr = [belongs(vstack((u,v,t)),kl_div_epi),
              geq(u,0),geq(v,0)]

    # Construct and return program
    p = program(minimize(t),
                constr,
                [u,v],
                name='kl_div')
    return p(x,y)



# Copyright © 2012-2025 Sergio Callegari, Federico Bizzarri
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <http://www.gnu.org/licenses/>.

"""Demo code for a ΔΣ modulator driving an electric motor at different
mechanical loads (and slips).

© 2012–2025 Sergio Callegari, Federico Bizzarri
All rights reserved.
"""

# Note. This code returns numeric results that are not fully consistent with
# those in the ICECS 2012 paper. Specifically, the SNR values
# evaluated via the NTF on the paper are better than those obtained by
# this code by a factor 2. In previous versions of the code, the results
# where identical due to a missing factor 2 in the quantization noise gain
# evaluation. Probably, this error slipped in because it makes the NTF-based
# results and the 'real' time-domain-simulation results more similar. In fact,
# the 'real' modulator behavior is better than predicted by the NTF model
# because in the specific example the quantization noise is not exactly white,
# but slightly blue (for some reason it gets more power at higher frequencies).



import numpy as np
import scipy as sp
import matplotlib.pyplot as plt
from pydsm.delsig import dbv, dbp, evalTF, synthesizeNTF, simulateDSM
from pydsm.ir import impulse_response
from pydsm.NTFdesign import quantization_noise_gain
from pydsm.NTFdesign.legacy import q0_from_filter_ir
from pydsm.NTFdesign.weighting import ntf_fir_from_q0


# Linearized motor model and default parameters
class Motor:
    P = 4       # Number of motor poles
    B = 25E-3   # Damping constant (dissipation due to windage and friction)
    J = 25E-3   # Moment of inertia
    Rs = 17.7   # Stator dissipative effects
    Rr = 13.8   # Rotor dissipative effects
    Ls = 459.2E-3   # Total 3-phase stator inductance
    Lr = 457.0E-3   # Total 3-phase rotor inductance
    Lm = 442.5E-3   # Magnetizing inductance
    fw = 50     # Nominal driving voltage angular frequency
    Vw = 320    # Nominal driving voltage peak amplitude

    def g0(self, sigma):
        return 1/self.Rs

    def tf_s(self, sigma, normalize=False):
        # Returns the linearized dynamical model of the motor
        # as a filter
        # sigma is slip coefficient
        b = np.array([sigma/self.Ls,
                      self.Rr/(self.Ls*self.Lr)])
        a = np.array([sigma*(1-self.Lm**2/(self.Ls*self.Lr)),
                      self.Rr/self.Lr+sigma*self.Rs/self.Ls,
                      self.Rr*self.Rs/(self.Lr*self.Ls)])
        if normalize:
            return (b/self.g0(sigma), a)
        else:
            return (b, a)

    def tf_z(self, sigma, fs, normalize=False):
        # Returns the linearized dynamical model of the motor
        # as a discrete time filter
        # sigma is slip coefficient
        # fs is the sample rate
        (btc, atc) = self.tf_s(sigma, normalize)
        return sp.signal.bilinear(btc, atc, fs)

motor = Motor()

# Motor data and Delta Sigma setup
fmot = 50.         # Nominal frequency of electrical drive
OSR = 1000         # Oversampling ratio
fphi = fmot*2*OSR  # Clock of Digital Delta Sigma Modulator
A = 190./320       # Normalized amplitude of the electric drive

# Approximation level (order) for the NTF
P = 8
# ... the same for the Delsig based NTF
DELSIG_P = 4

# Parameters to check the operation of the modulator in the time domain
Tstop = 2*OSR*100
Tstart = 2*OSR*80
# Intervals to make example plots of time domain data
T1p = Tstart
T2p = Tstart+4*OSR
dither_sigma = 1e-3

# The motor transfer functions to check (at different load/slip)
hs0043 = motor.tf_s(0.043)
hs02 = motor.tf_s(0.2)
hs06 = motor.tf_s(0.6)
hz0043 = motor.tf_z(0.043, fphi, True)
hz02 = motor.tf_z(0.2, fphi, True)
hz06 = motor.tf_z(0.6, fphi, True)

# Frequency axis for frequency domain plots (log scale)
fmin = fmot/OSR
fmax = fphi/2
ff = np.geomspace(fmin, fmax, 1024)

# Plot frequency response of the motor
yyz0043 = evalTF(hz0043, np.exp(2j*np.pi*ff/fphi))
yyz02 = evalTF(hz02, np.exp(2j*np.pi*ff/fphi))
yyz06 = evalTF(hz06, np.exp(2j*np.pi*ff/fphi))

plt.figure()
plt.semilogx(ff, dbv(np.abs(yyz0043)), 'b', label=r'$\sigma=0.043$')
plt.semilogx(ff, dbv(np.abs(yyz02)), 'r', label=r'$\sigma=0.2$')
plt.semilogx(ff, dbv(np.abs(yyz06)), 'g', label=r'$\sigma=0.6$')
plt.xlim(xmax=fmax)
plt.ylim(ymin=-50)
plt.xlabel('$f$ [Hz]')
plt.ylabel('[dB]')
plt.suptitle("Magnitude response of motor linearized model")
plt.legend()
plt.ion()
plt.show()
plt.ioff()

# Compute optimal and benchmark NTFs
hz0043_ir = impulse_response(hz0043)
hz02_ir = impulse_response(hz02)
hz06_ir = impulse_response(hz06)
q0_0043 = q0_from_filter_ir(P, hz0043_ir)
ntf0043 = ntf_fir_from_q0(q0_0043)
q0_02 = q0_from_filter_ir(P, hz02_ir)
ntf02 = ntf_fir_from_q0(q0_02)
q0_06 = q0_from_filter_ir(P, hz06_ir)
ntf06 = ntf_fir_from_q0(q0_06)
delsig_ntf = synthesizeNTF(DELSIG_P, OSR, 3, 1.5, 0)

# Plot the NTFs
nyz0043 = evalTF(ntf0043, np.exp(2j*np.pi*ff/fphi))
nyz02 = evalTF(ntf02, np.exp(2j*np.pi*ff/fphi))
nyz06 = evalTF(ntf06, np.exp(2j*np.pi*ff/fphi))
delsig_ny = evalTF(delsig_ntf, np.exp(2j*np.pi*ff/fphi))

plt.figure()
plt.semilogx(ff, dbv(np.abs(nyz0043)), 'b',
             label=r'opt @ $\sigma=0.043$')
plt.semilogx(ff, dbv(np.abs(nyz02)), 'r',
             label=r'opt @ $\sigma=0.2$')
plt.semilogx(ff, dbv(np.abs(nyz06)), 'g',
             label=r'opt @ $\sigma=0.6$')
plt.semilogx(ff, dbv(np.abs(delsig_ny)), 'y', label='benchmark')
plt.xlabel('$f$ [Hz]')
plt.ylabel('[dB]')
plt.suptitle("NTF magnitude response")
plt.xlim(xmax=fmax)
plt.ylim(-225, 5)
plt.legend(loc=4)

print("Computing the SNR in the various cases")
noise_power = 1./3  # Quantization noise is power is Delta^2/12
# low slip
n_pow_0043_0043 = noise_power*quantization_noise_gain(hz0043, ntf0043)
n_pow_0043_02 = noise_power*quantization_noise_gain(hz0043, ntf02)
n_pow_0043_06 = noise_power*quantization_noise_gain(hz0043, ntf06)
n_pow_0043_delsig = noise_power*quantization_noise_gain(hz0043, delsig_ntf)
# mid slip
n_pow_02_0043 = noise_power*quantization_noise_gain(hz02, ntf0043)
n_pow_02_02 = noise_power*quantization_noise_gain(hz02, ntf02)
n_pow_02_06 = noise_power*quantization_noise_gain(hz02, ntf06)
n_pow_02_delsig = noise_power*quantization_noise_gain(hz02, delsig_ntf)
# high slip
n_pow_06_0043 = noise_power*quantization_noise_gain(hz06, ntf0043)
n_pow_06_02 = noise_power*quantization_noise_gain(hz06, ntf02)
n_pow_06_06 = noise_power*quantization_noise_gain(hz06, ntf06)
n_pow_06_delsig = noise_power*quantization_noise_gain(hz06, delsig_ntf)

sig_pow = A**2/2  # Input signal power
# Output signal power at 3 slip conditions
sig_pow_0043 = sig_pow*np.abs(evalTF(hz0043, np.exp(2j*np.pi*fmot/fphi)))**2
sig_pow_02 = sig_pow*np.abs(evalTF(hz02, np.exp(2j*np.pi*fmot/fphi)))**2
sig_pow_06 = sig_pow*np.abs(evalTF(hz06, np.exp(2j*np.pi*fmot/fphi)))**2

print("Noise power and SNR for motor at slip=0.043")
print("ntf0043\tntf02\tntf06\tdelsig")
print("{:.4e}\t{:.4e}\t{:.4e}\t{:.4e}".format(
    n_pow_0043_0043, n_pow_0043_02, n_pow_0043_06, n_pow_0043_delsig))
print("{:2.2f}dB\t{:2.2f}dB\t{:2.2f}dB\t{:2.2f}dB".format(
    dbp(sig_pow_0043/n_pow_0043_0043), dbp(sig_pow_0043/n_pow_0043_02),
    dbp(sig_pow_0043/n_pow_0043_06), dbp(sig_pow_0043/n_pow_0043_delsig)))
print("Noise power and SNR for motor at slip=0.2")
print("ntf0043\tntf02\tntf06\tdelsig")
print("{:.4e}\t{:.4e}\t{:.4e}\t{:.4e}".format(
    n_pow_02_0043, n_pow_02_02, n_pow_02_06, n_pow_02_delsig))
print("{:2.2f}dB\t{:2.2f}dB\t{:2.2f}dB\t{:2.2f}dB".format(
    dbp(sig_pow_02/n_pow_02_0043), dbp(sig_pow_02/n_pow_02_02),
    dbp(sig_pow_02/n_pow_02_06), dbp(sig_pow_02/n_pow_02_delsig)))
print("Noise power and SNR for motor at slip=0.6")
print("ntf0043\tntf02\tntf06\tdelsig")
print("{:.4e}\t{:.4e}\t{:.4e}\t{:.4e}".format(
    n_pow_06_0043, n_pow_06_02, n_pow_06_06, n_pow_06_delsig))
print("{:2.2f}dB\t{:2.2f}dB\t{:2.2f}dB\t{:2.2f}dB".format(
    dbp(sig_pow_06/n_pow_06_0043), dbp(sig_pow_06/n_pow_06_02),
    dbp(sig_pow_06/n_pow_06_06), dbp(sig_pow_06/n_pow_06_delsig)))

# Repeating the tests in the time domain with the nonlinear modulator model
# and the linearized motor model
# Setting up DS simulation
tt = np.asarray(range(int(Tstop)))
uu = A*np.sin(2*np.pi*fmot/fphi*tt)
dither = np.random.randn(len(uu))*dither_sigma
uud = uu+dither
# Run the simulations
print("Simulating modulator with NTF optimized for sigma=0.043")
xx_opti0043 = simulateDSM(uud, ntf0043)[0]
print("Simulating modulator with NTF optimized for sigma=0.2")
xx_opti02 = simulateDSM(uud, ntf02)[0]
print("Simulating modulator with NTF optimized for sigma=0.6")
xx_opti06 = simulateDSM(uud, ntf06)[0]
print("Simulating modulator with benchmark NTF")
xx_delsig = simulateDSM(uud, delsig_ntf)[0]
print("Applying the filters from the linearized motor model")
print("Using the transfer function corresponding to sigma=0.043")
uud_filt0043 = sp.signal.lfilter(hz0043[0], hz0043[1], uud)
xx_opti_filt0043_0043 = sp.signal.lfilter(hz0043[0], hz0043[1], xx_opti0043)
xx_opti_filt0043_02 = sp.signal.lfilter(hz0043[0], hz0043[1], xx_opti02)
xx_opti_filt0043_06 = sp.signal.lfilter(hz0043[0], hz0043[1], xx_opti06)
xx_delsig_filt0043 = sp.signal.lfilter(hz0043[0], hz0043[1], xx_delsig)
print("Using the filter for sigma=0.2")
uud_filt02 = sp.signal.lfilter(hz02[0], hz02[1], uud)
xx_opti_filt02_0043 = sp.signal.lfilter(hz02[0], hz02[1], xx_opti0043)
xx_opti_filt02_02 = sp.signal.lfilter(hz02[0], hz02[1], xx_opti02)
xx_opti_filt02_06 = sp.signal.lfilter(hz02[0], hz02[1], xx_opti06)
xx_delsig_filt02 = sp.signal.lfilter(hz02[0], hz02[1], xx_delsig)
print("Using the filter for sigma=0.6")
uud_filt06 = sp.signal.lfilter(hz06[0], hz06[1], uud)
xx_opti_filt06_0043 = sp.signal.lfilter(hz06[0], hz06[1], xx_opti0043)
xx_opti_filt06_02 = sp.signal.lfilter(hz06[0], hz06[1], xx_opti02)
xx_opti_filt06_06 = sp.signal.lfilter(hz06[0], hz06[1], xx_opti06)
xx_delsig_filt06 = sp.signal.lfilter(hz06[0], hz06[1], xx_delsig)
print("Computing the merit factors")
noise_power_opti_0043_0043 = np.sum(
    (uud_filt0043[Tstart:]-xx_opti_filt0043_0043[Tstart:])**2)/(Tstop-Tstart)
noise_power_opti_0043_02 = np.sum(
    (uud_filt0043[Tstart:]-xx_opti_filt0043_02[Tstart:])**2)/(Tstop-Tstart)
noise_power_opti_0043_06 = np.sum(
    (uud_filt0043[Tstart:]-xx_opti_filt0043_06[Tstart:])**2)/(Tstop-Tstart)
noise_power_delsig_0043 = np.sum(
    (uud_filt0043[Tstart:]-xx_delsig_filt0043[Tstart:])**2)/(Tstop-Tstart)
noise_power_opti_02_0043 = np.sum(
    (uud_filt02[Tstart:]-xx_opti_filt02_0043[Tstart:])**2)/(Tstop-Tstart)
noise_power_opti_02_02 = np.sum(
    (uud_filt02[Tstart:]-xx_opti_filt02_02[Tstart:])**2)/(Tstop-Tstart)
noise_power_opti_02_06 = np.sum(
    (uud_filt02[Tstart:]-xx_opti_filt02_06[Tstart:])**2)/(Tstop-Tstart)
noise_power_delsig_02 = np.sum(
    (uud_filt02[Tstart:]-xx_delsig_filt02[Tstart:])**2)/(Tstop-Tstart)
noise_power_opti_06_0043 = np.sum(
    (uud_filt06[Tstart:]-xx_opti_filt06_0043[Tstart:])**2)/(Tstop-Tstart)
noise_power_opti_06_02 = np.sum(
    (uud_filt06[Tstart:]-xx_opti_filt06_02[Tstart:])**2)/(Tstop-Tstart)
noise_power_opti_06_06 = np.sum(
    (uud_filt06[Tstart:]-xx_opti_filt06_06[Tstart:])**2)/(Tstop-Tstart)
noise_power_delsig_06 = np.sum(
    (uud_filt06[Tstart:]-xx_delsig_filt06[Tstart:])**2)/(Tstop-Tstart)
sig_powx_0043 = np.sum(uud_filt0043[Tstart:]**2)/(Tstop-Tstart)
sig_powx_02 = np.sum(uud_filt02[Tstart:]**2)/(Tstop-Tstart)
sig_powx_06 = np.sum(uud_filt06[Tstart:]**2)/(Tstop-Tstart)
print("Noise power and SNR for motor at slip=0.043")
print("ntf0043\tntf02\tntf06\tdelsig")
print("{:.4e}\t{:.4e}\t{:.4e}\t{:.4e}".format(
    noise_power_opti_0043_0043, noise_power_opti_0043_02,
    noise_power_opti_0043_06, noise_power_delsig_0043))
print("{:2.2f}dB\t{:2.2f}dB\t{:2.2f}dB\t{:2.2f}dB".format(
    dbp(sig_powx_0043/noise_power_opti_0043_0043),
    dbp(sig_powx_0043/noise_power_opti_0043_02),
    dbp(sig_powx_0043/noise_power_opti_0043_06),
    dbp(sig_powx_0043/noise_power_delsig_0043)))
print("Noise power and SNR for motor at slip=0.2")
print("ntf0043\tntf02\tntf06\tdelsig")
print("{:.4e}\t{:.4e}\t{:.4e}\t{:.4e}".format(
    noise_power_opti_02_0043, noise_power_opti_02_02,
    noise_power_opti_02_06, noise_power_delsig_02))
print("{:2.2f}dB\t{:2.2f}dB\t{:2.2f}dB\t{:2.2f}dB".format(
    dbp(sig_powx_02/noise_power_opti_02_0043),
    dbp(sig_powx_02/noise_power_opti_02_02),
    dbp(sig_powx_02/noise_power_opti_02_06),
    dbp(sig_powx_02/noise_power_delsig_02)))
print("Noise power and SNR for motor at slip=0.6")
print("ntf0043\tntf02\tntf06\tdelsig")
print("{:.4e}\t{:.4e}\t{:.4e}\t{:.4e}".format(
    noise_power_opti_06_0043, noise_power_opti_06_02,
    noise_power_opti_06_06, noise_power_delsig_06))
print("{:2.2f}dB\t{:2.2f}dB\t{:2.2f}dB\t{:2.2f}dB".format(
    dbp(sig_powx_06/noise_power_opti_06_0043),
    dbp(sig_powx_06/noise_power_opti_06_02),
    dbp(sig_powx_06/noise_power_opti_06_06),
    dbp(sig_powx_06/noise_power_delsig_06)))

# Plot portions of the time domain behavior
plt.figure()
plt.suptitle("Current profile at sigma=0043")
plt.plot(tt[T1p:T2p], xx_opti_filt0043_0043[T1p:T2p], 'b', label='optimal NTF')
plt.plot(tt[T1p:T2p], xx_delsig_filt0043[T1p:T2p], 'r', label='benchmark NTF')
plt.plot(tt[T1p:T2p], uud_filt0043[T1p:T2p], 'g', label='ideal behavior')
plt.legend()
plt.xlabel('time')
plt.ylabel('normalized currents')

plt.show()



# Copyright © 2013–2025 Sergio Callegari, Federico Bizzarri
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <http://www.gnu.org/licenses/>.

"""Demo for ΔΣ modulator with assigned reconstruction filter.

© 2013–2025 Sergio Callegari, Federico Bizzarri
All rights reserved.
"""


import numpy as np
from scipy.signal import butter
from pydsm.delsig import evalTF, synthesizeNTF
from pydsm.delsig import dbv, dbp
from pydsm.NTFdesign import quantization_noise_gain
import matplotlib.pyplot as plt
from pydsm.NTFdesign import ntf_fir_weighting

# Signal specification
B = 1000.
OSR = 64
fphi = B*OSR*2
# Lee constraint
H_inf = 1.5
# FIR Order for optimal NTF
order = 10
# Order for reference NTF (DELSIG)
delsig_order = 4

# Define the reconstruction filter
# Cuts at 2B in order to have negligible attenuation in the signal band
hz = butter(1, B/fphi, btype='low', output='zpk')
hz_mag = lambda f: np.abs(evalTF(hz, np.exp(-2j*np.pi*f)))


print("... plotting filter magnitude response")
ff = np.logspace(np.log10(0.5E-5), np.log10(0.5), 1024).reshape((-1, 1))
vv = dbv(np.abs(hz_mag(ff)))
plt.figure()
plt.plot(ff, vv)
plt.xlim(1e-5, 1./2)
plt.ylim(-50, 5)
plt.xscale('log', base=10)
plt.gca().set_xticks([1E-5, 1./2])
plt.gca().set_xticklabels(['$10^{-5}$', r'$\frac{1}{2}$'])
plt.xlabel('$f$', x=1.)
plt.ylabel('$w(f)$ [dB]', y=0.9)
plt.grid(True, 'both')
plt.suptitle('Output filter magnitude response')
plt.tight_layout(rect=[0, 0, 1, 0.98])

print("... computing optimal NTF")
ntf_opti = ntf_fir_weighting(order, hz, H_inf=H_inf)
ntf_opti_mag = lambda f: np.abs(evalTF(ntf_opti, np.exp(-2j*np.pi*f)))

print("... computing reference NTF")
ntf_delsig = synthesizeNTF(delsig_order, OSR, 3, H_inf, 0)
ntf_delsig_mag = lambda f: np.abs(evalTF(ntf_delsig, np.exp(-2j*np.pi*f)))

print("... plotting optimal NTF amplitude response")
vv_mag = dbv(ntf_opti_mag(ff))
vv_ref = dbv(ntf_delsig_mag(ff))
plt.figure()
plt.plot(ff, vv_mag, 'b', label='proposed')
l = plt.plot(ff, vv_ref, 'r', label='reference')
plt.xlim(1e-5, 1./2)
plt.ylim(-100, 5)
plt.xscale('log', base=10)
plt.gca().set_xticks([1E-5, 1./2])
plt.gca().set_xticklabels(['$10^{-5}$', r'$\frac{1}{2}$'])
plt.xlabel('$f$', x=1.)
plt.ylabel((r'$\left|\mathit{NTF}\,\left(\mathrm{e}^{\mathrm{i} 2\pi f}' +
            r'\right)\right|$ [dB]'), y=0.59)
plt.grid(True, 'both')
plt.suptitle('Noise transfer function (magnitude)')
plt.legend(loc="lower right")
plt.tight_layout(rect=[0, 0, 1, 0.98])

print("... computing some interesting data")
p1 = quantization_noise_gain(ntf_opti, hz)
p2 = quantization_noise_gain(ntf_delsig, hz)

print("Overall noise power amplification/attenuation")
print("Proposed", p1, "(", dbp(p1), "dB)")
print("Reference", p2, "(", dbp(p2), "dB)")

plt.show()



# Copyright © 2013–2025 Sergio Callegari
# All rights reserved.

# This file is part of PyDSM.

# PyDSM is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# PyDSM is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with PyDSM.  If not, see <http://www.gnu.org/licenses/>.

"""Demo code to replicate the results shown in Figure 5 and in table 1,
of the paper:

S. Callegari 'Coding of Stereo Signals by a Single Digital ΔΣ Modulator'
IEEE 20th International Conference on Electronics, Circuits, and Systems
(ICECS 2013), pp. 589–592, Dec. 2013.

This code is about the proposed case, with a two band modulator and
multiplexing.

© 2013–2025 Sergio Callegari.
All rights reserved.
"""


import numpy as np
from scipy import signal
from pydsm.delsig import evalTF, dbv, dbp, simulateDSM, synthesizeNTF
import matplotlib.pyplot as plt
import matplotlib.mlab as mlab
from dualNTF import mirroredNTF
from extract_tone import find_tone

# Set up problem data
# There are two signals, each of them with its own band.
# In this code, the reference case where an independent modulator is
#
B1 = 20E3     # Bandwidth of first signal
B2 = 20E3     # Bandwidth of second signal
OSR = 64      # Oversampling ratio
A1 = 0.2      # Amplitude of first signal
A2 = 0.44     # Amplitude of the second signal
f1 = 1003     # Frequency used to test the first test signal with a tone
f2 = 3193     # Frequency used to test the second test signal with a tone

Amax = 0.34   # Max amplitude used for testing

# Clock frequency of modulator
fphi = 2*OSR*(B1+B2)

# Cut off frequency of reconstruction filters
# ...set twice as wide as they should to better see artifacts
brk1 = B1/fphi
brk2 = 0.5-B2/fphi

# Create NTF with DELSIG synthesizeNTF
ntf_a = synthesizeNTF(order=4, osr=2*OSR, opt=3, H_inf=np.sqrt(1.5))
ntf_dual_a = mirroredNTF(ntf_a)

# Prepare frequency grid
ff_log = np.logspace(np.log10(0.5E-5), np.log10(0.5), 4096)
ff_lin = np.linspace(0, 0.5, 1024)

# Compute magnitude responses of NTF for plotting
ntf_a_mag = lambda f: np.abs(evalTF(ntf_a, np.exp(-2j*np.pi*f)))
vv_ntf_a_log = ntf_a_mag(ff_log)
vv_ntf_a_lin = ntf_a_mag(ff_lin)
ntf_dual_a_mag = lambda f: np.abs(evalTF(ntf_dual_a, np.exp(-2j*np.pi*f)))
vv_ntf_dual_a_log = ntf_dual_a_mag(ff_log)
vv_ntf_dual_a_lin = ntf_dual_a_mag(ff_lin)

# Design output filter
hz = signal.butter(4, 2*brk1, btype='low')

# Compute magnitude responses of output filter for plotting
hz_mag = lambda f: np.abs(evalTF(hz, np.exp(-2j*np.pi*f)))
vv_hz_log = hz_mag(ff_log)
vv_hz_lin = hz_mag(ff_lin)

print("Plotting the noise transfer function...")

# Plot magnitude responses with log scale - same as Fig. 5a
plt.figure()
plt.plot(ff_log, dbv(vv_ntf_dual_a_log))
plt.xlim(5E-5, 0.5)
plt.ylim(-119.99, 5)
plt.xscale('log', base=10)
plt.xlabel(r'$\hat f$', x=1.)
plt.ylabel(r'magnitude [dB]')
plt.gca().set_xticks([1E-5, 0.5])
plt.gca().set_xticklabels(['$10^{-5}$', r'$\frac{1}{2}$'])
plt.suptitle('NTF mag response - log scale')

# Plot magnitude responses with linear scale - same as Fig. 5b
plt.figure()
plt.plot(ff_lin, dbv(vv_ntf_dual_a_lin))
plt.xlim(0, 0.5)
plt.ylim(-100, 5)
plt.xlabel(r'$\hat f$', x=1.)
plt.ylabel(r'magnitude [dB]')
plt.gca().set_xticks([0, 1./2])
plt.gca().set_xticklabels(['$0$', r'$\frac{1}{2}$'])
plt.suptitle('NTF mag response - linear scale')

print("Doing time domain simulations...")

# Define test time for time domain simulations
Tstop = 200*fphi/1000  # approximately 200 periods at 1kHz
tt = np.arange(Tstop, dtype=int)
print("Test time:", len(tt), 'points')

# Generate signals for time domain simulations
print("Generate input signals ", end=" ")
s1 = A1*np.sin(2*np.pi*f1*tt/fphi)
print("* ", end="")
s2 = A2*np.sin(2*np.pi*f2*tt/fphi)
print("* ", end="")
clk = 2*(tt % 2)-1
print("# ", end="")
s2r = s2*clk
print("*")
dither_sigma = 1e-9  # Use a little dither
dither = np.random.randn(len(tt))*dither_sigma

# Do the simulation
print("Simulating DSM", end=" ")
xx1 = simulateDSM(s1+s2r+dither, ntf_dual_a)[0]
print("* ", end=""),
xx2 = xx1*clk
print("*")

# Plot a fragment of the modulator output - same as Fig. 5c
plt.figure()
plt.plot(tt[100:150], xx1[100:150], drawstyle='steps-post')
plt.xlabel(r'$n$', x=1.)
plt.ylabel(r'$x(nT)$', x=1.)
plt.ylim(-1.2, 1.2)
plt.suptitle('Fragment of modulator output - 1st signal')

# Reconstruct the signals
print("Applying a low pass filter")
yy1 = signal.lfilter(hz[0], hz[1], xx1)
yy2 = signal.lfilter(hz[0], hz[1], xx2)

# Plot fragment of first reconstructed signal - same as Figs. 5e, 5f
Tdisp_start = 10*fphi/1000
Tdisp_stop = 15*fphi/1000
ttp = np.arange(Tdisp_start, Tdisp_stop, dtype=int)
plt.figure()
plt.plot(ttp/fphi*1000, yy1[ttp], 'b', label='1st signal')
plt.plot(ttp/fphi*1000, yy2[ttp], 'r', label='2nd signal')
plt.xlabel(r'$t$ [ms]', x=1.)
plt.ylabel(r'$\hat u_1(t)$')
plt.suptitle('Fragment of reconstructed output')
plt.legend()

# Estimate some spectra
print("Compute some spectra (", end="")
Tsp_start = int(2*fphi/1000)
Tsp_stop = int(199*fphi/1000)
NFFT = 4096*64
(psd, freqs) = mlab.psd((xx1)[Tsp_start:Tsp_stop], Fs=fphi, NFFT=NFFT,
                        noverlap=NFFT//2, scale_by_freq=True)
print('@modulator, ', end="")
(psd1, freqs1) = mlab.psd((yy1)[Tsp_start:Tsp_stop], Fs=fphi, NFFT=NFFT,
                          noverlap=NFFT//2, scale_by_freq=True)
print('@1st signal, ', end="")
(psd2, freqs2) = mlab.psd((yy2)[Tsp_start:Tsp_stop], Fs=fphi, NFFT=NFFT,
                          noverlap=NFFT//2, scale_by_freq=True)
print('@2nd signal)')

print("Check overall power from psd")
print(" (should be 1): ", end="")
df = freqs[1]-freqs[0]
print(np.sum(psd)*df)

# Plot PDS of signal at modulator output
# This is the same as in Fig. 5b
# Try to express quantities in dBm (referred to 1 mW over a 50 Ohm resistor)
plt.figure()
plt.plot(freqs, dbp(psd)+30-10*np.log10(50))
plt.xlim(0, fphi/2)
plt.xlabel(r'$f$ [MHz]', x=1.)
plt.ylabel(r'PDS [dBm/Hz]')
plt.gca().set_xticks([0, fphi/2])
plt.gca().set_xticklabels(['$0$', r'$2.56$'])
plt.suptitle('PSD at modulator output')

# Plot PDS of first reconstructed signal
plt.figure()
plt.plot(freqs1, dbp(psd1))
plt.xscale('log', base=10)
plt.suptitle('PSD of reconstructed signal - 1st signal')

# Plot PDS of second reconstructed signal
plt.figure()
plt.plot(freqs2, dbp(psd2))
plt.xscale('log', base=10)
plt.suptitle('PSD of reconstructed signal - 2nd signal')

# Check modulator output when tone is subtracted
tone_xx1 = find_tone(xx1, f1, fphi, start=Tsp_start, N=Tsp_stop-Tsp_start,
                     output='polar')
tone_xx2 = find_tone(xx2, f2, fphi, start=Tsp_start, N=Tsp_stop-Tsp_start,
                     output='polar')
# Remove the tone, quick version
xx1q = xx1-tone_xx1[0]*np.cos(2*np.pi*f1/fphi*tt-tone_xx1[1])
xx2q = xx2-tone_xx2[0]*np.cos(2*np.pi*f2/fphi*tt-tone_xx2[1])
(psd1q, freqs1q) = mlab.psd(xx1q[Tsp_start:Tsp_stop], Fs=fphi, NFFT=NFFT,
                            noverlap=NFFT//2, scale_by_freq=True)
(psd2q, freqs2q) = mlab.psd(xx2q[Tsp_start:Tsp_stop], Fs=fphi, NFFT=NFFT,
                            noverlap=NFFT//2, scale_by_freq=True)

# Plot PDS of quantization noise for channel 1 at modulator output
# This is the same as Fig. 5g
plt.figure()
plt.plot(freqs1q/1000, dbp(psd1q)+30-10*np.log10(50))
plt.xscale('log', base=10)
plt.xlim(freqs1q[0]/1000, freqs1q[-1]/1000)
plt.xlabel(r'$f$ [kHz]', x=1.)
plt.ylabel(r'q. noise #1 [dBm/Hz]')
plt.suptitle('PSD of quantization noise - 1st signal')

# Plot PDS of quantization noise for channel 2 at modulator output
# This is the same as Fig. 5h
plt.figure()
plt.plot(freqs2q/1000, dbp(psd2q)+30-10*np.log10(50))
plt.xscale('log', base=10)
plt.xlim(freqs2q[0]/1000, freqs2q[-1]/1000)
plt.ylabel(r'q. noise #2 [dBm/Hz]')
plt.xlabel(r'$f$ [kHz]', x=1.)
plt.suptitle('PSD of quantization noise - 2nd signal')

# Compute merit factors referred to the modulator output
print('Compute merit factors for delta sigma sequences')
df1q = freqs1q[1]-freqs1q[0]
df2q = freqs2q[1]-freqs2q[0]
ifm1 = int(np.round((B1-freqs1q[0])/df1q)+1)
ifm2 = int(np.round((B2-freqs2q[0])/df2q)+1)
ibqnp1 = np.sum(psd1q[0:ifm1])*df1q
ibqnp2 = np.sum(psd2q[0:ifm1])*df2q
print('Noise floors', dbp(ibqnp1)+30-dbp(50), dbp(ibqnp2)+30-dbp(50), 'dBm')
print('SNRs', dbp(0.5*A1**2/ibqnp1), dbp(0.5*A2**2/ibqnp2), 'dB')
print('SNRmax', dbp(0.5*Amax**2/ibqnp1), dbp(0.5*Amax**2/ibqnp2), 'dB')

# Compute merit factors referred to the reconstructed signals
tone1 = find_tone(yy1, f1, fphi, start=Tsp_start, N=Tsp_stop-Tsp_start,
                  output='polar')
tone2 = find_tone(yy2, f2, fphi, start=Tsp_start, N=Tsp_stop-Tsp_start,
                  output='polar')
# Remove the tone
yy1q = yy1-tone1[0]*np.cos(2*np.pi*f1/fphi*tt-tone1[1])
yy2q = yy2-tone2[0]*np.cos(2*np.pi*f2/fphi*tt-tone2[1])
print('Signals at filter output:')
print('tone 1:', tone1)
print('tone 2:', tone2)

(psd1yq, freqs1yq) = mlab.psd(yy1q[Tsp_start:Tsp_stop], Fs=fphi, NFFT=NFFT,
                              noverlap=NFFT//2, scale_by_freq=True)
(psd2yq, freqs2yq) = mlab.psd(yy2q[Tsp_start:Tsp_stop], Fs=fphi, NFFT=NFFT,
                              noverlap=NFFT//2, scale_by_freq=True)
df1yq = freqs1yq[1]-freqs1yq[0]
df2yq = freqs2yq[1]-freqs2yq[0]
ibqnpy1 = np.sum(psd1yq)*df1yq
ibqnpy2 = np.sum(psd2yq)*df2yq
print('Noise floors', dbp(ibqnpy1)+30-dbp(50), dbp(ibqnpy2)+30-dbp(50), 'dB')
print('SNRs', dbp(0.5*tone1[0]**2/ibqnpy1), dbp(0.5*tone2[0]**2/ibqnpy2), 'dB')

# Plot PDS of quantization noise for channel 1 at filter output
# The same as in Fig. 5g, after filtering
plt.figure()
plt.plot(freqs1q/1000, dbp(psd1yq)+30-10*np.log10(50))
plt.xscale('log', base=10)
plt.xlim(freqs1q[0]/1000, freqs1q[-1]/1000)
plt.xlabel(r'$f$ [kHz]', x=1.)
plt.ylabel(r'q. noise #1 [dBm/Hz]')
plt.suptitle('PSD of quantization noise - 1st signal - filtered')

# Plot PDS of quantization noise for channel 2 at filter output
# The same as in Fig. 5h, after filtering
plt.figure()
plt.plot(freqs2q/1000, dbp(psd2yq)+30-10*np.log10(50))
plt.xscale('log', base=10)
plt.xlim(freqs2q[0]/1000, freqs2q[-1]/1000)
plt.ylabel(r'q. noise #2 [dBm/Hz]')
plt.xlabel(r'$f$ [kHz]', x=1.)
plt.suptitle('PSD of quantization noise - 2nd signal - filtered')

# Look for crosstalk
tonex1 = find_tone(yy2, f1, fphi, start=Tsp_start, N=Tsp_stop-Tsp_start,
                   output='polar')
tonex2 = find_tone(yy1, f2, fphi, start=Tsp_start, N=Tsp_stop-Tsp_start,
                   output='polar')
print('Crosstalk at filter output:')
print('tone 1:', tonex1, dbp(tonex1[0]**2)+30-dbp(50), 'dBm')
print('tone 2:', tonex2, dbp(tonex2[0]**2)+30-dbp(50), 'dBm')

plt.show()
